import { database as stockDatabase } from "./AggregateDatabase/Stock/lmdbx.js";
import { database as calendarDatabase } from "./AggregateDatabase/Calendar/optiondb-lmdbx.js";
import type { CalendarKey } from "./AggregateDatabase/Calendar/interfaces.js";
import { nextDate } from "./lib/utils/nextDate.js";

export type BacktestInput = {
  symbol: string;
  startDate: string;
  endDate: string;
  /** Between 0 and 1. The frequency that similar calendars have historically ended (i.e. within the last hour) at a higher price than the current calendar's price. */
  historicalProbabilityOfSuccess?: number;
  initialBuyingPower?: number;
};

export async function backtest({
  symbol,
  startDate,
  endDate,
  historicalProbabilityOfSuccess = 0.8,
  initialBuyingPower = 2000,
}: BacktestInput) {
  let buyingPower = initialBuyingPower;
  const portfolio = new Set<CalendarKey>();
  // for each day:
  for (
    let date = startDate, didBuyCalendar = false;
    date <= endDate;
    date = nextDate(date), didBuyCalendar = false
  ) {
    const calendars = await calendarDatabase.getCalendars({
      key: { symbol },
      date,
    });
    const stockAggregates = await stockDatabase.getAggregates({
      key: { symbol },
      date,
    });
    // for each minute of that day for which we have a stock candlestick:
    for (const stockAggregate of stockAggregates) {
      // console.log("Current Time:", new Date(stockAggregate.tsStart));
      // filter-out calendars that are far-from-the-money (5%)
      const calendarsNearTheMoney = calendars.filter(
        ({ strike }) =>
          Math.abs((stockAggregate.open - strike) / stockAggregate.open) < 0.05
      );
      console.log(
        "Current Date:",
        new Intl.DateTimeFormat("en-US", {
          timeZone: "America/New_York",
          dateStyle: "full",
          timeStyle: "long",
        }).format(new Date(stockAggregate.tsStart)),
        ";",
        `${calendarsNearTheMoney.length} Calendars Near The Money`
      );
      // for each relevant calendar on that day:
      for (const calendar of calendarsNearTheMoney) {
        const strikePercentageFromTheMoney = Math.abs(
          (stockAggregate.open - calendar.strike) / stockAggregate.open
        );
        /** Number of days between the back and front expiration dates. */
        const calendarSpanInDays =
          (new Date(calendar.backExpirationDate).valueOf() -
            new Date(calendar.frontExpirationDate).valueOf()) /
          (1000 * 60 * 60 * 24);
        const targetCalendarPrice =
          await calendarDatabase.getTargetPriceByProbability({
            symbol,
            calendarSpan: calendarSpanInDays,
            strikePercentageFromTheMoney,
            historicalProbabilityOfSuccess,
          });
        const calendarAggregateAtCurrentTime =
          await calendarDatabase.getAggregate({
            key: {
              ...calendar,
            },
            tsStart: stockAggregate.tsStart,
          });
        // if there exists a matching calendar candlestick for the current minute:
        if (calendarAggregateAtCurrentTime) {
          // if the current candlestick is a good price (i.e. less than the target price):
          const minCalendarPriceInCandlestick = Math.min(
            calendarAggregateAtCurrentTime.open,
            calendarAggregateAtCurrentTime.close
          );
          if (
            minCalendarPriceInCandlestick < targetCalendarPrice &&
            minCalendarPriceInCandlestick >
              0.07 /* sometimes the calendar price is zero or negative, which is of course impossible; some institution got a good deal */
          ) {
            // if we can afford to buy the calendar:
            if (buyingPower > minCalendarPriceInCandlestick) {
              // buy the calendar, and continue to the next day:
              portfolio.add(calendar);
              buyingPower = buyingPower - minCalendarPriceInCandlestick * 100;
              console.log(
                "Bought",
                calendar,
                "for",
                minCalendarPriceInCandlestick * 100,
                "...$",
                buyingPower,
                "left"
              );
              didBuyCalendar = true;
            }
          }
        }
        if (didBuyCalendar) {
          break;
        }
      }
      if (didBuyCalendar) {
        break;
      }
    }

    // for each calendar in portfolio, if today is the last day, close the position:
    for (const calendar of portfolio.values()) {
      if (calendar.frontExpirationDate === date) {
        const calendarClosingPrice = await calendarDatabase.getClosingPrice({
          key: {
            ...calendar,
          },
        });
        portfolio.delete(calendar);
        buyingPower = buyingPower + calendarClosingPrice * 100;
        console.log(
          "Sold",
          calendar,
          "for",
          calendarClosingPrice,
          "...$",
          buyingPower,
          "left"
        );
      }
    }
  }

  return {
    endingBuyingPower: buyingPower,
    portfolio: Array.from(portfolio.values()),
  };
}
