import { publicProcedure, router, RpcType } from "./trpc.js";
import { query } from "./lib/clickhouse.js";
import { createHTTPHandler } from "@trpc/server/adapters/standalone";
import cors from "cors";
import {
  Object as ObjectT,
  String as StringT,
  Number as NumberT,
} from "@sinclair/typebox";
import { createServer } from "node:http";
import { Env } from "@humanwhocodes/env";
import StockPriceChart from "./StockPriceChart.js";
import SimilarCalendarPriceChart from "./SimilarCalendarPriceChart.js";
import CalendarExitPriceChart from "./CalendarExitPriceChart.js";
import CalendarCharacteristicsForm from "./CalendarCharacteristicsForm.js";

const env = new Env();

const LISTEN_PORT = env.get("LISTEN_PORT", 3005);

export const getOpensForUnderlying = publicProcedure
  .input(
    RpcType(
      ObjectT({
        underlying: StringT({ maxLength: 5 }),
      })
    )
  )
  .query(async (opts) => {
    const { underlying } = opts.input;
    return await query<{ x: number; y: number }>(
      `
		SELECT
			toUnixTimestamp(tsStart) as x,
			open as y
		FROM stock_aggregates
		WHERE symbol = '${underlying}'
		ORDER BY tsStart ASC
	`,
      "JSONEachRow"
    );
  });

export const getOpensForOptionContract = publicProcedure
  .input(
    RpcType(
      ObjectT({
        underlying: StringT({ maxLength: 5 }),
        expirationDate: StringT(),
        strike: NumberT(),
      })
    )
  )
  .query(async (opts) => {
    const { underlying, expirationDate, strike } = opts.input;
    return await query<{ x: number; y: number }>(
      `
		SELECT
			toUnixTimestamp(tsStart) as x,
			open as y
		FROM option_contract_aggregates
		WHERE symbol = '${underlying}'
		AND expirationDate = '${expirationDate}'
		AND strike = ${strike}
		AND type = 'call'
		ORDER BY tsStart ASC
	`,
      "JSONEachRow"
    );
  });

const appRouter = router({
  CalendarCharacteristicsForm,
  StockPriceChart,
  SimilarCalendarPriceChart,
  CalendarExitPriceChart,

  getOpensForUnderlying,
  getOpensForOptionContract,
});

// Export type router type signature,
// NOT the router itself.
export type AppRouter = typeof appRouter;

const handler = createHTTPHandler({
  middleware: cors(),
  router: appRouter,
  createContext() {
    return {};
  },
});

const server = createServer((req, res) => {
  if (req.url.startsWith("/healthz")) {
    res.statusCode = 200;
    res.end("OK");
  } else {
    handler(req, res);
  }
});

server.listen(Number.parseInt(LISTEN_PORT));
